import torch.nn as nn
import torch
import torch.nn.functional as F
from torch.utils.data import TensorDataset
from torch.optim.lr_scheduler import ExponentialLR
import numpy as np
from sklearn.model_selection import KFold, StratifiedKFold
import matplotlib.pyplot as plt
import os

# Here we define our self.net as a class
class Net(nn.Module):

    def __init__(self, input_dim, hidden_dim, output_dim, num_lstm_layers=1):
        super(Net, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.output_dim = output_dim
        self.num_lstm_layers = num_lstm_layers
        # Define the LSTM layer
        self.lstm = nn.LSTM(self.input_dim, self.hidden_dim, self.num_lstm_layers, dropout=0.3)
        # nn.init.normal_(self.lstm.weight, mean=0, std=1.0)
        self.dropout = nn.Dropout(p=0.3)
        # Define the output layer
        self.linear = nn.Linear(self.hidden_dim, self.output_dim)
        nn.init.normal_(self.linear.weight, mean=0, std=1.0)

    def forward(self, input):
        # Forward pass through LSTM layer
        # shape of lstm_out: [input_size, batch_size, hidden_dim]
        # shape of self.hidden: (a, b), where a and b both
        # have shape (num_layers, batch_size, hidden_dim).
        lstm_out, self.hidden = self.lstm(input, self.hidden)
        flattened_lstm_out = lstm_out[-1].view(-1, self.hidden_dim)
        # linear_ip = F.softmax(flattened_lstm_out)
        linear_ip = flattened_lstm_out
        dropout_op = self.dropout(linear_ip)

        # Only take the output from the final timetep
        # Can pass on the entirety of lstm_out to the next layer if it is a seq2seq prediction
        y_pred = self.linear(dropout_op)
        # y_pred = F.softmax(y_pred, dim=0)
        return y_pred

class LSTM:
    def __init__(self, input_dim=9, hidden_dim=20, output_dim=4, save_path= "../data/models/", learning_rate = 0.0005, weight_decay=0.005, num_lstm_layers=1):
        self.save_path=save_path
        if torch.cuda.is_available():
            self.device = torch.device("cuda:0") # Uncomment this to run on GPU'
            print("DEVICE IS CUDA")
        else:
            self.device = torch.device("cpu")
            print("DEVICE IS CPU :()")
        self.net = Net(input_dim, hidden_dim, output_dim).to(self.device)
        print(self.net)
        self.loss_function = nn.CrossEntropyLoss()
        # self.loss_function = nn.MSELoss()
        self.optimizer = torch.optim.Adam(self.net.parameters(), weight_decay=weight_decay, lr=learning_rate)
        # self.scheduler = ExponentialLR(self.optimizer, gamma=0.9)
        self.num_lstm_layers = num_lstm_layers


    def init_hidden(self, batch_size):
        # This is what we'll initialise our hidden state as
        return (torch.zeros(self.num_lstm_layers, batch_size, self.net.hidden_dim).to(device=self.device),
                torch.zeros(self.num_lstm_layers, batch_size, self.net.hidden_dim).to(device=self.device))

    def save(self, epoch, loss):
        PATH = self.save_path + "model-" + str(epoch)
        torch.save({
            'epoch': epoch,
            'model_state_dict': self.net.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'loss': loss
            }, PATH)

    def train(self, X, y, num_epochs = 150, callback=None):

        mean_loss_hist = []
        mean_train_accuracy_hist = []
        mean_validation_accuracy_hist = []

        std_loss_hist = []
        std_train_accuracy_hist = []
        std_validation_accuracy_hist = []

        fig = plt.figure()
        ax1 = fig.add_subplot(221)
        ax1.set_title('Training Loss', fontsize=15)
        ax2 = fig.add_subplot(222, sharex=ax1)
        ax2.set_ylim([0, 1])
        ax2.set_title('Training Accuracy', fontsize=15)
        ax3 = fig.add_subplot(223, sharex=ax1)
        ax3.set_title('validation Accuracy', fontsize=15)
        ax3.set_ylim([0, 1])
        ax4 = fig.add_subplot(224, sharex=ax1)
        ax4.set_title('Test Accuracy', fontsize=15)
        ax4.set_ylim([0, 1])
        # kfold = KFold(n_splits=6)
        # kfold = StratifiedKFold(n_splits=6)
        kfold = StratifiedKFold(n_splits=6)
        for t in range(num_epochs):
            epoch_loss = []
            epoch_train_accuracies = []
            epoch_validation_accuracies = []
            for train_index, validation_index in kfold.split(X, y):
                x_train_fold = X[train_index]
                y_train_fold = y[train_index]
                x_validation_fold = X[validation_index]
                y_validation_fold = y[validation_index]

                # print(x_train_fold.shape, y_train_fold.shape)
                # print(x_validation_fold.shape, y_validation_fold.shape)

                Xvalidation = torch.from_numpy(x_validation_fold).float().to(device=self.device)
                Xvalidation = Xvalidation.transpose(0,1)
                yvalidation = torch.from_numpy(y_validation_fold).float().type(torch.LongTensor).to(device=self.device)

                Xtrain = torch.from_numpy(x_train_fold).float().to(device=self.device)
                Xtrain = Xtrain.transpose(0,1)
                ytrain = torch.from_numpy(y_train_fold).float().type(torch.LongTensor).to(device=self.device)
                train_batch_size = len(train_index)
                validation_batch_size = len(validation_index)
                self.net.train()
                self.net.zero_grad()
                # Initialise hidden state
                # Don't do this if you want your LSTM to be stateful
                self.net.hidden = self.init_hidden(train_batch_size)
                # Xt = batch x time_steps x features
                # Forward pass
                y_pred = self.net(Xtrain)
                loss = self.loss_function(y_pred, ytrain)
                epoch_loss.append(loss.item())
                # Zero out gradient, else they will accumulate between epochs
                self.optimizer.zero_grad()

                # Backward pass
                loss.backward()

                # Update parameters
                self.optimizer.step()

                train_accuracy = self.evaluate(Xtrain, ytrain)
                epoch_train_accuracies.append(train_accuracy.item())
                # mean_train_accuracy_hist.append(train_accuracy.item())
                validation_accuracy = self.evaluate(Xvalidation, yvalidation)
                epoch_validation_accuracies.append(validation_accuracy.item())

                # mean_validation_accuracy_hist.append(validation_accuracy.item())
                print("Epoch ", t, "Loss : ", loss.item(), "Train Accuracy : ", train_accuracy.item(), "validation Accuracy : ", validation_accuracy.item() )
            # self.scheduler.step()
            mean_loss_hist.append(np.mean(epoch_loss))
            mean_train_accuracy_hist.append(np.mean(epoch_train_accuracies))
            mean_validation_accuracy_hist.append(np.mean(epoch_validation_accuracies))

            std_loss_hist.append(np.std(epoch_loss))
            std_train_accuracy_hist.append(np.std(epoch_train_accuracies))
            std_validation_accuracy_hist.append(np.std(epoch_validation_accuracies))

            np.save(os.path.join(self.save_path, "mean_loss_hist"), mean_loss_hist)
            np.save(os.path.join(self.save_path, "std_loss_hist"), std_loss_hist)

            np.save(os.path.join(self.save_path, "mean_train_accuracy_hist"), mean_train_accuracy_hist)
            np.save(os.path.join(self.save_path, "std_train_accuracy_hist"), std_train_accuracy_hist)

            np.save(os.path.join(self.save_path, "mean_validation_accuracy_hist"), mean_validation_accuracy_hist)
            np.save(os.path.join(self.save_path, "std_validation_accuracy_hist"), std_validation_accuracy_hist)
            # print("len mean loss", len(mean_loss_hist))
            # print("len STD", len(std_loss_hist))
            # print("Size x",np.arange(0, len(mean_loss_hist)).shape)
            if callback != None:
                callback(self, t, ax4, self.save_path)

            if t%10 ==0:
                self.save(t, loss)
                PATH = self.save_path + "performance-" + str(t) + ".svg"
                plt.savefig(PATH)
                ax1.clear()
                ax1.set_title('Training Loss')
                ax1.set(xlabel='# of Epochs', ylabel='Loss')
                ax1.plot(np.arange(0, len(mean_loss_hist)), mean_loss_hist)
                ax1.fill_between(np.arange(0, len(mean_loss_hist)), np.asarray(mean_loss_hist)-np.asarray(std_loss_hist), np.asarray(mean_loss_hist)+np.asarray(std_loss_hist), alpha=0.5)

                ax2.clear()
                ax2.set_ylim([0, 1])
                ax2.set_title('Training Accuracy')
                ax2.set(xlabel='# of Epochs', ylabel='Accuracy')
                ax2.plot(np.arange(0, len(mean_train_accuracy_hist)), mean_train_accuracy_hist)
                ax2.fill_between(np.arange(0, len(mean_train_accuracy_hist)),  np.asarray(mean_train_accuracy_hist)-np.asarray(std_train_accuracy_hist), np.asarray(mean_train_accuracy_hist)+np.asarray(std_train_accuracy_hist), alpha=0.5)

                ax3.clear()
                ax3.set_title('validation Accuracy')
                ax3.set(xlabel='# of Epochs', ylabel='Accuracy')
                ax3.set_ylim([0, 1])
                ax3.plot(np.arange(0, len(mean_validation_accuracy_hist)), mean_validation_accuracy_hist)
                ax3.fill_between(np.arange(0, len(mean_validation_accuracy_hist)), np.asarray(mean_validation_accuracy_hist)-np.asarray(std_validation_accuracy_hist), np.asarray(mean_validation_accuracy_hist)+np.asarray(std_validation_accuracy_hist), alpha=0.5)
                mng = plt.get_current_fig_manager()
                mng.window.showMaximized()
                plt.pause(0.005)

        self.save(t, loss)
        ax1.clear()
        ax1.set_title('Training Loss')
        ax1.set(xlabel='# of Epochs', ylabel='Loss')
        ax1.plot(np.arange(0, len(mean_loss_hist)), mean_loss_hist)
        ax1.fill_between(np.arange(0, len(mean_loss_hist)), np.asarray(mean_loss_hist)-np.asarray(std_loss_hist), np.asarray(mean_loss_hist)+np.asarray(std_loss_hist), alpha=0.5)

        ax2.clear()
        ax2.set_ylim([0, 1])
        ax2.set_title('Training Accuracy')
        ax2.set(xlabel='# of Epochs', ylabel='Accuracy')
        ax2.plot(np.arange(0, len(mean_train_accuracy_hist)), mean_train_accuracy_hist)
        ax2.fill_between(np.arange(0, len(mean_train_accuracy_hist)),  np.asarray(mean_train_accuracy_hist)-np.asarray(std_train_accuracy_hist), np.asarray(mean_train_accuracy_hist)+np.asarray(std_train_accuracy_hist), alpha=0.5)

        ax3.clear()
        ax3.set_title('validation Accuracy')
        ax3.set(xlabel='# of Epochs', ylabel='Accuracy')
        ax3.set_ylim([0, 1])
        ax3.plot(np.arange(0, len(mean_validation_accuracy_hist)), mean_validation_accuracy_hist)
        ax3.fill_between(np.arange(0, len(mean_validation_accuracy_hist)), np.asarray(mean_validation_accuracy_hist)-np.asarray(std_validation_accuracy_hist), np.asarray(mean_validation_accuracy_hist)+np.asarray(std_validation_accuracy_hist), alpha=0.5)
        mng = plt.get_current_fig_manager()
        mng.window.showMaximized()
        plt.pause(0.005)
        if callback != None:
            callback(self, t, ax4, self.save_path)
        PATH = self.save_path + "performance-" + str(t) + ".svg"
        plt.savefig(PATH)
        plt.show()

    def evaluate(self, X, y):
        self.net.eval()
        predictions = self.predict(X)
        accuracy = self.getAccuracy(predictions, y)
        return accuracy

    def predict(self, X):
        predictions = self.predict_probs(X)
        predictions = torch.argmax(predictions, axis=1)
        return predictions

    def predict_probs(self, X):
        batch_size = X.shape[1]
        self.net.hidden = self.init_hidden(batch_size)
        predictions = self.net(X)
        # Next line wasn't there before
        predictions = F.softmax(predictions, dim=0)
        return predictions

    def getAccuracy(self, predictions, labels):
        loss = torch.eq(predictions, labels)
        total_right =torch.sum(loss).float()
        accuracy = total_right/float(labels.shape[0])
        return accuracy
    # def predict(self):
    #
