import scipy.io as sio
import numpy as np
import os
import argparse
import fnmatch
import pandas as pd
from sklearn.model_selection import train_test_split
from ML.LSTM import LSTM
import torch
import matplotlib.pyplot as plt
from sklearn import preprocessing
from sklearn.metrics import roc_curve, auc
import pickle
from statistics import mean

# import matplotlib
# size = 15
# params = {'legend.fontsize': 'large',
#           # 'figure.figsize': (20,8),
#           'axes.labelsize': size,
#           'axes.titlesize': size,
#           'xtick.labelsize': size*0.75,
#           'ytick.labelsize': size*0.75,
#           'axes.titlepad': 25,
#           'font.size': size}
# matplotlib.rcParams.update(params)

mean_test_accuracy_hist = []
test_X = None
test_y = None

def organizeTrialData(data_dict, signal_idx):
    sequences = list()
    labels = list()
    for subject_diagnoses in data_dict.keys():
        for current_trial_data in data_dict[subject_diagnoses]:
            current_trial_data = current_trial_data[ ~np.isnan(current_trial_data).any(axis=1) ]
            current_trial_data = current_trial_data[:, signal_idx]
            sequences.append(current_trial_data)
            labels.append(subject_diagnoses)

    return sequences, labels

def pruneSequences(sequences, targets):
    len_sequences = []
    unskipped_sequences = []
    unskipped_targets = []

    for (one_seq, one_target) in zip(sequences, targets):
        if len(one_seq) <= 105:
            continue
        unskipped_sequences.append(one_seq)
        unskipped_targets.append(one_target)
        len_sequences.append(len(one_seq))
    min_length = min(len_sequences)
    print("Min Trial Length", min_length)
    print("Max Trial Length", max(len_sequences))
    print("Average Trial Length", mean(len_sequences))
    for i in range(len(unskipped_sequences)):
        old_array = unskipped_sequences[i]
        new_array = old_array[:min_length, :]
        unskipped_sequences[i] = new_array
    print("Num targets", len(unskipped_targets))
    return unskipped_sequences, unskipped_targets

def callback(lstm, t, axis, save_path):
    test_accuracy = lstm.evaluate(test_X, test_y)
    mean_test_accuracy_hist.append(test_accuracy.item())
    np.save(os.path.join(save_path, "mean_test_accuracy_hist"), mean_test_accuracy_hist)

    axis.clear()
    axis.set_title('Test Accuracy')
    axis.set(xlabel='# of Epochs', ylabel='Accuracy')
    axis.set_ylim([0, 1])
    axis.plot(mean_test_accuracy_hist)
    print("\t  Test Accuracy : ", test_accuracy.item() )

def getSubsetInput(seq, start_idx=3, end_idx=6):
    return seq[:, :, start_idx:end_idx]


def compute_one_v_rest_roc(lstm, label_mappings, save_path):
    global test_y
    fpr = dict()
    tpr = dict()
    roc_auc = dict()

    lstm.net.eval()
    y_score = lstm.predict_probs(test_X).cpu()
    test_y = test_y.detach().cpu().numpy()
    y_score = y_score.detach().cpu().numpy()

    for ndd_class in label_mappings.keys():
        i = label_mappings[ndd_class]
        fpr[i], tpr[i], _ = roc_curve(test_y, y_score[:, i], pos_label=i)
        roc_auc[i] = auc(fpr[i], tpr[i])
        # print("\tfpr for ", i, fpr[i])
        # print("\ttpr for ", i, tpr[i])
        print("\tROC for ", ndd_class, ":", i, roc_auc[i])
        print("--------------------")

    with open(os.path.join(save_path, "fpr.pkl"), "wb") as f:
        pickle.dump(fpr, f)

    with open(os.path.join(save_path, "tpr.pkl"), "wb") as f:
        pickle.dump(tpr, f)

    with open(os.path.join(save_path, "roc_auc.pkl"), "wb") as f:
        pickle.dump(roc_auc, f)
    plot_individual_rocs(label_mappings, fpr, tpr, roc_auc, save_path)
    plot_single_roc(label_mappings, fpr, tpr, roc_auc, save_path)


def plot_individual_rocs(label_mappings, fpr, tpr, roc_auc, save_path):
    for ndd_class in label_mappings.keys():
        i = label_mappings[ndd_class]
        plt.figure()
        plt.plot(fpr[i], tpr[i], label='ROC curve for %s (area = %0.2f)' % (ndd_class, roc_auc[i]), linewidth=4)
        plt.plot([0, 1], [0, 1], 'k--', linewidth=4)
        plt.xlim([-0.05, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate', fontsize=20)
        plt.ylabel('True Positive Rate', fontsize=20)
        plt.title('Receiver operating characteristic for %s vs All' % ndd_class, fontsize=30)
        plt.legend(loc="lower right", prop={"size":20})
        plt.gca().tick_params(axis='both', labelsize=15)

        mng = plt.get_current_fig_manager()
        mng.window.showMaximized()
        plt.pause(0.005)
        fig_path = os.path.join(save_path, "roc_" + str(ndd_class) + ".svg")
        plt.savefig(fig_path)
        plt.show()


def plot_single_roc(label_mappings, fpr, tpr, roc_auc, save_path):
    plt.figure()
    colors = ['blue', 'red', 'green', 'black']
    for ndd_class in label_mappings.keys():
        i = label_mappings[ndd_class]
        color = colors[i]

        plt.plot(fpr[i], tpr[i], color=color, label='ROC curve for %s (area = %0.2f)' % (ndd_class, roc_auc[i]), linewidth=4)
    plt.plot([0, 1], [0, 1], 'k--', linewidth=4)
    plt.xlim([-0.05, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate', fontsize=20)
    plt.ylabel('True Positive Rate', fontsize=20)
    plt.title('Receiver operating characteristic for multi-class data', fontsize=30)
    plt.legend(loc="lower right", prop={"size":20})
    plt.gca().tick_params(axis='both', labelsize=15)

    mng = plt.get_current_fig_manager()
    mng.window.showMaximized()
    plt.pause(0.005)
    fig_path = os.path.join(save_path, "roc_All.svg")
    plt.savefig(fig_path)

    plt.show()
    # for class in label_mappings
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Train LSTM.')
    input_signals = ["linear_accel", "rpy_dot", "rpy", "linear_accel_and_rpy_dot", "linear_accel_and_rpy", "rpy_dot_and_rpy", "all"]
    parser.add_argument('input_signal', type=str, default="rpy_dot_and_rpy", choices=input_signals,
                        help='Select what input signal to use. Choices include '+', '.join(input_signals))
    parser.add_argument('--ip_file_path', help='Path to data input')
    parser.add_argument('--op_file_path', help='Path to data output')
    parser.add_argument('--test_ratio', type=float, default=0.35)
    parser.add_argument("--num_epochs", type=int, default=500)

    args = parser.parse_args()
    print(args)
    # ip_file_path = "../data/apdm/data/all_data-control+category.npy"
    # ip_file_path = "../data/apdm/data/all_data.npy"
    ip_file_path = "../../Data_analyzed/all_trial_data.npy"
    if args.ip_file_path != None:
        ip_file_path = args.ip_file_path
    # op_file_path = "../data/apdm/models/"
    op_file_path = "../data/xsens/models/"
    if args.op_file_path != None:
        op_file_path = args.op_file_path

    if args.input_signal == "linear_accel":
        sigal_idx = [0, 1, 2]
    if args.input_signal == "rpy_dot":
        sigal_idx = [3, 4, 5]
    if args.input_signal == "rpy":
        sigal_idx = [6, 7, 8]
    if args.input_signal == "linear_accel_and_rpy_dot":
        sigal_idx = [0, 1, 2, 3, 4, 5]
    if args.input_signal == "linear_accel_and_rpy":
        sigal_idx = [0, 1, 2, 6, 7, 8]
    if args.input_signal == "rpy_dot_and_rpy":
        sigal_idx = [3, 4, 5, 6, 7, 8]
    if args.input_signal == "all":
        sigal_idx = [0, 1, 2, 3, 4, 5, 6, 7, 8]
    with open(os.path.join(op_file_path, "settings.txt"), "w") as f:
        f.write(args.__str__())
    all_data = np.load(ip_file_path, allow_pickle=True)
    all_data = all_data.item()
    sequences, targets = organizeTrialData(all_data, sigal_idx)
    new_seq, targets = pruneSequences(sequences, targets)
    print("Num seq", len(new_seq))
    print("Num targets", len(targets))
    final_seq = np.stack(new_seq)
    print(final_seq.shape)

    num_outputs = len(list(set(targets)))
    print("single shape ", sequences[0].shape)
    print("num outputs ", num_outputs)
    # print(targets)
    targets = np.array(targets)
    le = preprocessing.LabelEncoder()
    le.fit(targets)
    labels = le.transform(targets)
    label_mappings = dict(zip(le.classes_, le.transform(le.classes_)))


    # final_seq = getSubsetInput(final_seq)
    # print(final_seq.shape)

    X_train, X_test, y_train, y_test = train_test_split(final_seq, labels, stratify=labels, test_size=args.test_ratio)
    print("Final Seq shape", final_seq.shape)
    print("X_test shape", X_test.shape)
    print("y_train", y_train)
    print("y_test", y_test)

    np.save(os.path.join(op_file_path, "X_train"), X_train)
    np.save(os.path.join(op_file_path, "y_train"), y_train)

    np.save(os.path.join(op_file_path, "X_test"), X_test)
    np.save(os.path.join(op_file_path, "y_test"), y_test)

    num_features = final_seq.shape[2]
    seq_len = final_seq.shape[1]
    if torch.cuda.is_available():
        device = torch.device("cuda:0") # Uncomment this to run on GPU'
        print("DEVICE IS CUDA")
    else:
        device = torch.device("cpu")
        print("DEVICE IS CPU :()")
    test_X = torch.from_numpy(X_test).float().to(device=device)
    test_X = test_X.transpose(0,1)
    test_y = torch.from_numpy(y_test).float().type(torch.LongTensor).to(device=device)
    print(num_features)
    print(num_outputs)

    lstm = LSTM(input_dim = num_features, output_dim=num_outputs, save_path = op_file_path, hidden_dim=30, learning_rate = 0.005, weight_decay=0.005)
    # lstm.train(X_train, y_train, X_test, y_test,batch_size=len(y_train), num_epochs=1000)
    lstm.train(X_train, y_train, num_epochs=args.num_epochs, callback=callback)
    lstm.save(args.num_epochs + 1, 0)

    compute_one_v_rest_roc(lstm, label_mappings, op_file_path )
